package se.petterhansson.gp;

import java.util.Stack;

import se.petterhansson.sevengp.gp.alphabet.SevenGPNode;

public class Individual<T extends GPNode> {
	private final T root;
	private final int treeSize;
	
	/**
	 * Creates a completely random individual.
	 * @param depth Max depth.
	 * @param terminateDepth A depth where it becomes OK to insert terminals. Inverted so that the
	 * largest value is at root. Use this param to make sure tree is of at least a certain height.
	 * @param source
	 */
	public Individual(int depth, int terminateDepth, RandomSource<T> source) {
		if (depth <= 0) {
			throw new RuntimeException("Bad depth");
		}
		
		root = randomTree(depth, terminateDepth, source);
		treeSize = root.getTreeSize();
	}
	
	/**
	 * Creates a mutant from a parent.
	 * @param depth Max depth of random tree inserted.
	 * @param parent
	 * @param source
	 */
	@SuppressWarnings("unchecked")
	public Individual(int depth, Individual<T> parent, RandomSource<T> source) {
		T parentClone;
		try {
			parentClone = (T) parent.root.clone();
		}
		catch (CloneNotSupportedException e) {
			throw new RuntimeException("Bad clone");
		}
		
		T insertionNode = selectRandomBranch(parentClone);
		T randomTree = randomTree(depth, depth, source);
		GPNode insertionParent = insertionNode.getParent();
		
		if (insertionParent == null) {
			root = randomTree;
		}
		else {
			root = parentClone;
			
			int index = insertionParent.getChildIndex(insertionNode);
			assert (index >= 0);
			
			insertionParent.removeChild(index);
			insertionParent.setChild(index, randomTree);
		}
		
		treeSize = root.getTreeSize();
	}

	/**
	 * Creates a cross over of two parents.
	 * @param parent1
	 * @param parent2
	 */
	@SuppressWarnings("unchecked")
	public Individual(Individual<T> parent1, Individual<T> parent2) {
		T parent1Clone;
		try {
			parent1Clone = (T) parent1.root.clone();
		}
		catch (CloneNotSupportedException e) {
			throw new RuntimeException("Bad clone");
		}
		
		T insertionNode = selectRandomBranch(parent1Clone);
		T extractionNode;
		try {
			extractionNode = (T) selectRandomBranch(parent2.getRoot()).clone();
		}
		catch (CloneNotSupportedException e) {
			throw new RuntimeException("Bad clone");
		}
		
		GPNode insertionParent = insertionNode.getParent();
		
		if (insertionParent == null) {
			root = extractionNode;
		}
		else {
			root = parent1Clone;
			
			int index = insertionParent.getChildIndex(insertionNode);
			assert (index >= 0);
			
			insertionParent.removeChild(index);
			insertionParent.setChild(index, extractionNode);
		}
		
		treeSize = root.getTreeSize();
	}
	
	/**
	 * Creates an individual from a genome string.
	 * @param genome
	 * @param factory
	 */
	public Individual(String genome, GPNodeFactory<T> factory) {
		root = null;
		treeSize = 0;
	}
	
	/**
	 * Creates a clone of an individual.
	 * @param parent
	 */
	@SuppressWarnings("unchecked")
	public Individual(Individual<SevenGPNode> parent) {
		T parentClone;
		try {
			parentClone = (T) parent.root.clone();
		}
		catch (CloneNotSupportedException e) {
			throw new RuntimeException("Bad clone");
		}
		
		root = parentClone;
		treeSize = parent.treeSize;
	}

	private T randomTree(int depth, int terminateDepth, RandomSource<T> source) {
		if (depth == 1) {
			return source.createRandomTerminal();
		}
		else
		{
			boolean mayTerminate = depth <= terminateDepth;
			T node = source.createRandomNode(mayTerminate);
			
			for (int i = 0, n = node.getArity(); i < n; ++i) {
				node.setChild(i, randomTree(depth - 1, terminateDepth, source));
			}
			
			return node;
		}
	}
	
	private T selectRandomBranch(T node) {
		Stack<T> stack = new Stack<T>();
		insertRandomChild(node, stack);
		int index = (int) (Math.random() * stack.size());
		
		return stack.get(index);
	}
	
	@SuppressWarnings("unchecked")
	private void insertRandomChild(T node, Stack<T> stack) {
		stack.push(node);
		if (node.getArity() > 0)
		{
			int index = (int) (Math.random() * node.getArity());
			
			insertRandomChild((T) node.getChild(index), stack);
		}
	}
	
	public String toString() {
		StringBuilder str = new StringBuilder();
		printTree(root, str);
		
		return str.toString();
	}

	private void printTree(GPNode node, StringBuilder str) {
		if (node == null) {
			return;
		}
		
		str.append(node.getPrefix());
		str.append(' ');
		
		for (int i = 0, n = node.getArity(); i < n; ++i) {
			printTree(node.getChild(i), str);
		}
	}

	public T getRoot() {
		return root;
	}

	public int getTreeSize() {
		return treeSize;
	}
}
