package se.petterhansson.sevengp;

import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;

import se.petterhansson.sevengp.game.SevenGame;
import se.petterhansson.sevengp.game.SevenPlayer;
import se.petterhansson.sevengp.game.SevenPlayerFactory;
import se.petterhansson.sevengp.game.TestSevenPlayer;
import se.petterhansson.sevengp.gp.SevenGPPlayer;
import se.petterhansson.sevengp.gp.SevenRandomSource;
import se.petterhansson.util.MultiOutputStream;

public class SevenGP {
	private final SevenRandomSource randomSource = new SevenRandomSource();
	private final ArrayList<Specimen> population = new ArrayList<Specimen>();
	private int generation;
	private int nextId;
	
	public static void main(String[] args) throws IOException {
		//no user interface. I expect to run this on its own anyway
		
		FileOutputStream file = new FileOutputStream("computer.txt");
		MultiOutputStream stream = new MultiOutputStream(new OutputStream[] { file, System.out} );
		
		for (int k = 2; k <= 6; ++k) {
			System.out.println("Generating AI for " + k + " players");
			
			SevenGP gp = new SevenGP();
			gp.setNumberOfPlayers(k);
			gp.addRandomPopulation((k - 1) * 1000);
			
			gp.testPerformance(5);
			gp.export(stream, "ai[" + k + "]['easy']", 50, 20);
			gp.runGenerations(1, 200);
			gp.export(stream, "ai[" + k + "]['medium']", 100, 20);
			gp.runGenerations(10, 200);
			gp.export(stream, "ai[" + k + "]['hard']", 250, 20);
			//gp.runGenerations(100, 200);
			//gp.export(stream, "ai[" + k + "]['expert']", 500, 20);
		}
		
		file.close();
	}
	
	// Permanent code starts here =============================================
	
	private int numPlayers;

	/**
	 * Adds a number of random specimen to the population.
	 * @param n
	 */
	public void addRandomPopulation(int n) {
		for (int i = 0; i < n; ++i) {
			population.add(new Specimen(nextId++, randomSource));
		}
	}

	/**
	 * Sets the number of players that shall be used for GP.
	 * @param n
	 */
	public void setNumberOfPlayers(int n) {
		numPlayers = n;
	}

	/**
	 * Runs the game for a number of generations.
	 * @param n Number of generations.
	 * @param g Games per generation.
	 */
	public void runGenerations(int n, int g) {
		for (int i = 0; i < n; ++i) {
			generation += 1;
			System.out.println("Running generation " + generation + "...");
			
			runSimulation(g);
			
			sortAccordingToFitness();
			testPerformance(5);
			
			applyGenetics();
		}
	}
	
	/**
	 * Performs cloning, mutation and cross over to entire population.
	 * Presumes that population is ordered by descending fitness.
	 */
	private void applyGenetics() {
		ArrayList<Specimen> newPopulation = new ArrayList<Specimen>();
		
		//decide proportions
		int remaining = population.size();
		int cloneSplit = remaining / 10;
		remaining -= cloneSplit;
		int alterMagnitudeSplit = remaining / 5;
		remaining -= alterMagnitudeSplit;
		int mutateSplit = remaining / 2;
		remaining -= mutateSplit;
		int crossSplit = remaining;
		
		//clone (in reality just transfer)
		for (int i = 0; i < cloneSplit; ++i) {
			newPopulation.add(population.get(i));
		}
		
		//alter magnitudes of constants
		for (int i = 0; i < alterMagnitudeSplit; ++i) {
			newPopulation.add(new Specimen(nextId++, population.get(i), true));
		}
		
		//mutate
		for (int i = 0; i < mutateSplit; ++i) {
			newPopulation.add(new Specimen(nextId++, population.get(i), randomSource));
		}
			
		//cross
		for (int i = 0; i < crossSplit; ++i) {
			int r = (int) (Math.random() * population.size());
			newPopulation.add(new Specimen(nextId++, population.get(i), population.get(r)));
		}
		
		population.clear();
		population.addAll(newPopulation);
	}

	/**
	 * Sorts population according to win ratio.
	 */
	private void sortAccordingToFitness() {
		Collections.sort(population, new Comparator<Specimen>() {
			@Override
			public int compare(Specimen a, Specimen b) {
				return (int) Math.signum(b.getFitness() - a.getFitness());
			}
		});
	}
	
	/**
	 * Resets specimen scores.
	 */
	private void clearScores() {
		for (Specimen specimen : population) {
			specimen.games = 0;
			specimen.score = 0;
		}
	}
	
	/**
	 * Runs the simulation.
	 */
	private void runSimulation(int g) {
		//essentially, try let each specimen compete against all other specimen twice (might need to reduce later)
		
		clearScores();
		
		ArrayList<Specimen> lineup = new ArrayList<Specimen>();
		
		for (int i = 0, n = population.size(); i < n; ++i) {
			Specimen me = population.get(i);
			
			lineup.clear();
			lineup.add(me);
			int gamesPlayed = 0;
			
			for (int j = 0; j < g; ++j) {
				Specimen opponent = population.get((int) (Math.random() * n));
				if (opponent == me) {
					--j;
					continue;
				}
				
				lineup.add(opponent);
				
				if (lineup.size() == numPlayers) {
					//we have enough players to start a game
					int winner = compete(lineup);
					for (int k = 0; k < numPlayers; ++k) {
						Specimen s = lineup.get(k);
						s.games += 1;
						
						if (winner < 0) {
							s.score -= 10; //we don't like bots that waste time
						}
						else if (winner == k)
						{
							lineup.get(winner).score += 1;
						}
					}
					
					lineup.clear();
					lineup.add(me);
					gamesPlayed += 1;
				}
			}
		}
	}
	
	/**
	 * Measures performance of n best scoring specimen against a random test player.
	 * Mainly useful to verify that GP is working and that there is real improvement in play.
	 * @param n
	 */
	public void testPerformance(int n) {
		System.out.println("Will test " + n + " specimen");
		for (int i = 0; i < n; ++i) {
			Specimen specimen = population.get(i);
			System.out.println(specimen.sessionId + ": " + specimen.getWinRatio() + " - " + specimen.games);
			System.out.println(specimen.play.toString());
		}
		
		clearScores();
		
		ArrayList<Specimen> lineup = new ArrayList<Specimen>();
		for (int i = 0; i < numPlayers - 1; ++i) {
			lineup.add(null); //yes, really, signifies default test player
		}
		
		for (int i = 0; i < n; ++i) {
			Specimen specimen = population.get(i);
			lineup.add(specimen);
			
			int trials = 100;
			int wins = 0;
			for (int j = 0; j < trials; ++j) {
				if (compete(lineup) == numPlayers - 1) {
					wins += 1;
				}
			}
			
			System.out.println("Specimen " + specimen.sessionId + " won " + ((float) wins / trials) + " of its games");
			
			lineup.remove(numPlayers - 1);
		}
	}
	
	/**
	 * The function that performs the actual game...
	 * @param specimen List of specimen. Any null specimen will be interpreted as test players.
	 * @return
	 */
	private int compete(final List<Specimen> specimen) {
		SevenGame game = new SevenGame(numPlayers, new SevenPlayerFactory() {
			@Override
			public SevenPlayer createSevenPlayer(SevenGame game, int index, int[] cards) {
				Specimen s = specimen.get(index);
				if (s == null) {
					return new TestSevenPlayer(game, index, cards, false);
				}
				
				return new SevenGPPlayer(game, index, cards, s.play, s.send);
			}
		});
		
		while (!game.isFinished()) {
			//System.out.println(game.toString());
			game.nextPlayer();
			
			if (game.getNumRounds() >= 100) {
				break;
			}
		}
		
		if (game.getWinner() != null) {
			return game.getWinner().getPlayerIndex();
		}
		
		return -1;
	}
	
	/**
	 * 
	 * @param os Stream to print to.
	 * @param storagePrefix Assignable Lua location (variable name or table member reference). Expected to
	 * 	be a table so that it can be indexed from 1 to n.
	 * @param g Number of games used in trials.
	 * @param n Number of survivors to be exported.
	 */
	public void export(OutputStream os, String storagePrefix, int g, int n) {
		assert(n <= population.size());
		
		System.out.println("Running final trials prior to export...");
		runSimulation(g);
		
		System.out.println("Exporting...");
		sortAccordingToFitness();
		
		PrintStream stream = new PrintStream(os);
		for (int i = 0; i < n; ++i) {
			stream.print(storagePrefix);
			stream.print('[');
			stream.print(i + 1);
			stream.print("] = ");
			
			Specimen specimen = population.get(i);
			stream.println(specimen.getLuaTable(""));
		}
	}
}
