package se.petterhansson.sevengp.game;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Iterator;
import java.util.Stack;

public class SevenGame implements Iterable<SevenPlayer> {
	private ArrayList<SevenPlayer> players;
	private int startPlayer;
	private int currentPlayer;
	private int rounds;
	private SevenPlayer winner;
	private int[] lowest = new int[4];
	private int[] highest = new int[4];
	private int[] owner = new int[52];
	
	public SevenGame(int nPlayers, SevenPlayerFactory factory) {
		for (int i = 0; i < 4; ++i) {
			lowest[i] = -1;
			highest[i] = -1;
		}
		
		//create and shuffle deck
		Stack<Integer> deck = new Stack<Integer>();
		
		for (int i = 0; i < 13 * 4; ++i) {
			deck.push(i);
		}
		
		Collections.shuffle(deck);
		
		//create players
		players = new ArrayList<SevenPlayer>();
		
		currentPlayer = -1;
		
		for (int i = 0; i < nPlayers; ++i) {
			int cardsLeft = deck.size();
			int nCards = cardsLeft / (nPlayers - i);
			int[] cards = new int[nCards];
			
			for (int j = 0; j < nCards; ++j) {
				cards[j] = deck.pop();
				owner[cards[j]] = i;
				
				if (Cards.getRank(cards[j]) == 7 - Cards.RANK_ZERO &&
						Cards.getSuit(cards[j]) == Cards.SUIT_DIAMOND) {
					assert (currentPlayer == -1);
					startPlayer = i;
					currentPlayer = i;
				}
			}
			
			players.add(factory.createSevenPlayer(this, i, cards));
		}
		
		assert (currentPlayer != -1);
	}
	
	public void nextPlayer() {
		SevenPlayer player = players.get(currentPlayer);
		
		if (winCheck(player)) {
			return;
		}
		
		//find leftmost (clock-wise) player with at least one card
		SevenPlayer leftPlayer = null;
		for (int i = currentPlayer; ;) {
			i = (i - 1) % players.size();
			leftPlayer = players.get(i < 0? i + players.size() : i);
			
			if (leftPlayer.countCards() > 0) {
				break;
			}
		}
		
		//check if active player can play a card
		boolean canPlay = false;
		for (int card : player) {
			if (canCardBePlayed(card)) {
				canPlay = true;
			}
		}
		
		//take decision on what the player can do
		if (canPlay) {
			int card = player.playCard();
			playCard(card);
		}
		else
		{
			int card = leftPlayer.sendCard();
			player.receiveCard(card);
		}
		
		currentPlayer = (currentPlayer + 1) % players.size();
		if (currentPlayer == startPlayer) {
			rounds += 1;
		}
	}
	
	public int getNumRounds() {
		return rounds;
	}
	
	private boolean winCheck(SevenPlayer player) {
		if (winner != null) {
			return true;
		}
		
		if (player.countCards() == 0 || canPlayerExit(player)) {
			winner = player;
			return true;
		}
		
		return false;
	}

	private boolean canPlayerExit(SevenPlayer player) {
		//back up lowest and highest
		int[] copiedLowest = Arrays.copyOf(lowest, 4);
		int[] copiedHighest = Arrays.copyOf(highest, 4);
		boolean[] played = new boolean[52];
		
		int nPlayed = 0;
		boolean change = true;
		for (int n = player.countCards(); nPlayed < n && change; ) {
			change = false;
			
			//check all player cards
			for (int card : player) {
				if (!played[card] && canCardBePlayed(card)) {
					playCard(card);
					
					played[card] = true;
					nPlayed += 1;
					change = true;
				}
			}
		}
		
		if (nPlayed == player.countCards()) {
			return true;
		}
		
		//restore
		for (int i = 0; i < 4; ++i) {
			lowest[i] = copiedLowest[i];
			highest[i] = copiedHighest[i];
		}
		
		return false;
	}

	public boolean canCardBePlayed(int card) {
		int suit = Cards.getSuit(card);
		int rank = Cards.getRank(card);
		
		if (lowest[suit] == -1) {
			return rank == 7 - Cards.RANK_ZERO;
		}
		
		int d = Math.min(Math.abs(lowest[suit] - rank), Math.abs(highest[suit] - rank));
		
		return d == 1;
	}
	
	private void playCard(int card) {
		assert (card != -1);
		assert (canCardBePlayed(card));
		
		owner[card] = -1;
		
		int suit = Cards.getSuit(card);
		int rank = Cards.getRank(card);
		
		if (lowest[suit] == -1) {
			assert (rank == 7 - Cards.RANK_ZERO);
			lowest[suit] = rank;
			highest[suit] = rank;
		}
		else
		{
			lowest[suit] = Math.min(lowest[suit], rank);
			highest[suit] = Math.max(highest[suit], rank);
		}
	}
	
	public int getLowestInSuit(int suit) {
		return lowest[suit];
	}
	
	public int getHighestInSuit(int suit) {
		return highest[suit];
	}
	
	public boolean isFinished() {
		return winner != null;
	}
	
	public SevenPlayer getWinner() {
		return winner;
	}

	@Override
	public Iterator<SevenPlayer> iterator() {
		return players.iterator();
	}

	public int getCurrentPlayer() {
		return currentPlayer;
	}
	
	public String toString() {
		StringBuilder str = new StringBuilder();
		
		str.append("Lowest and highest:\n");
		for (int i = 0; i < 4; ++i) {
			str.append(lowest[i]);
			str.append(' ');
			str.append(highest[i]);
			str.append('\n');
		}
		
		str.append('\n');
		
		for (int i = 0; i < players.size(); ++i) {
			str.append('\n');
			str.append(players.get(i).toString());
			str.append('\n');
		}
		
		return str.toString();
	}

	public int getCardOwner(int card) {
		return owner[card];
	}
}
